# coding=utf-8
import base64
import hashlib
import json
import logging
import os
import random
import socket
import sys

import click
import requests
from terminaltables import PorcelainTable
from six.moves.urllib import parse
from mali_commands.config import Config
from mali_commands.utilities.tables import format_json_data, dict_to_csv

base_url = '_ah/api/missinglink/v1/'


def print_as_json(result):
    click.echo(json.dumps(result))


def print_as_bad_json(result):
    click.echo(result)


def create_validate_length(min_length, max_length):
    def validate_length(ctx, option, value):
        if value is None:
            return value

        if len(value) > max_length:
            raise click.BadParameter("needs to be shorter than %s characters" % max_length)

        if len(value) < min_length:
            raise click.BadParameter("needs to be longer than %s characters" % min_length)

        return value

    return validate_length


def build_auth0_url(auth0):
    return '{}.auth0.com'.format(auth0)


def add_to_data_if_not_none(data, val, key):
    if val is not None:
        data[key] = val


def get_data_and_remove(kwargs, key):
    key = key.lower()
    result = kwargs.get(key)

    if key in kwargs:
        del kwargs[key]

    return result


def base64url(b):
    return bytearray(base64.b64encode(b).decode('ascii').replace('=', '').replace('+', '-').replace('/', '_'), 'ascii')


def sha256(s):
    h = hashlib.sha256()
    h.update(s)
    return h.digest()


def urljoin(*args):
    base = args[0]
    for u in args[1:]:
        base = parse.urljoin(base, u)

    return base


# noinspection PyUnusedLocal
def handle_api(ctx_or_config, http_method, method_url, data=None):
    config = ctx_or_config if isinstance(ctx_or_config, Config) else ctx_or_config.config

    if config.id_token is None:
        click.echo('Please run: "mali auth init" to setup authorization', err=True)
        exit(1)

    url = urljoin(config.api_host, base_url, method_url)

    id_token = config.id_token
    result = None
    for retries in range(3):
        headers = {'Authorization': 'Bearer {}'.format(id_token)}
        r = http_method(url, headers=headers, json=data)

        if r.status_code == 401:
            id_token = update_token(config)
            continue

        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as ex:
            try:
                error_message = ex.response.json().get('error', {}).get('message')
            except ValueError:
                error_message = str(ex)

            if error_message:
                click.echo('\n' + error_message, err=True)
                exit(1)
                return

            raise

        result = r.json()
        break

    if result is None:
        click.echo('failed to refresh the token, rerun auth init again', err=True)
        exit(1)
        return

    return result


def get_prefix_section(config_prefix, section):
    return '%s-%s' % (config_prefix, section) if config_prefix else section


def update_token(config):
    r = requests.post('https://{}/delegation'.format(build_auth0_url(config.auth0)), json={
        'client_id': config.client_id,
        'grant_type': "urn:ietf:params:oauth:grant-type:jwt-bearer",
        'scope': 'openid offline_access user_external_id org orgs email picture name given_name user_metadata',
        'refresh_token': config.refresh_token,
    })

    r.raise_for_status()

    data = r.json()

    config.set(get_prefix_section(config.config_prefix, 'token'), 'id_token', data['id_token'])
    config.save()

    return data['id_token']


class LocalWebServer(object):
    ports = [27017, 41589, 55130, 6617, 1566, 17404, 44288, 6948, 15950, 48216, 3318, 34449, 22082, 22845, 37862, 59304]

    def __init__(self):
        self.app = self.create_app()
        self.code = None
        self.srv = None

        self.disable_logging()
        self.start_server()

    @classmethod
    def disable_logging(cls):
        log = logging.getLogger('werkzeug')
        log.setLevel(logging.ERROR)

    def create_app(self):
        from flask import Flask, request, redirect
        app = Flask(__name__)

        @app.route('/console/auth/init')
        def auth_init():
            self.code = request.args.get('code')
            return redirect("/console/auth/finish")

        @app.route('/console/auth/finish')
        def auth_finish():
            self.srv.shutdown_signal = True
            return "You can go now back to the console"

        return app

    def run(self):
        self.srv.serve_forever()

    @property
    def redirect_uri(self):
        return 'http://localhost:%s/console/auth/init' % self.srv.port

    def start_server(self):
        from werkzeug.serving import make_server

        ports = self.ports[:]
        random.shuffle(ports)

        for port in ports:
            try:
                self.srv = make_server('127.0.0.1', port, self.app)
                break
            except (OSError, socket.error):
                continue

        if self.srv is None:
            click.echo("can't create local server for authentication, consider using the --disable-webserver flag")


def wait_for_input():
    if sys.version_info >= (3, 0):
        code = input('Enter the token ')
    else:
        # noinspection PyUnresolvedReferences
        code = raw_input('Enter the token ')

    return code


def auth0_oauth_token(auth0_url, code, verifier, client_id, redirect_uri):
    params = {
        'code': code,
        'code_verifier': verifier,
        'client_id': client_id,
        'grant_type': 'authorization_code',
        'redirect_uri': redirect_uri,
    }

    r = requests.post('https://{}/oauth/token'.format(auth0_url), json=params)

    r.raise_for_status()

    return r.json()


# noinspection PyUnusedLocal
def pixy_flow(ctx):
    import webbrowser

    verifier = base64url(os.urandom(32))
    verifier_challenge = base64url(sha256(verifier))

    verifier = verifier.decode('ascii')
    verifier_challenge = verifier_challenge.decode('ascii')

    app = LocalWebServer() if ctx.local_web_server else None

    redirect_uri = app.redirect_uri if app is not None else '{host}.firebaseapp.com/console/auth/success'.format(host=ctx.host)

    query = {
        'response_type': 'code',
        'scope': 'openid offline_access user_external_id org orgs email picture name given_name user_metadata',
        'client_id': ctx.client_id,
        'redirect_uri': redirect_uri,
        'code_challenge': verifier_challenge,
        'code_challenge_method': 'S256'
    }

    authorize_url = 'https://{}/authorize?{}'.format(build_auth0_url(ctx.auth0), parse.urlencode(query))

    click.echo("If the browser doesn't open enter this URL manually\n%s\n" % authorize_url)

    webbrowser.open(authorize_url)

    if app is not None:
        app.run()
        code = app.code
    else:
        code = wait_for_input()

    data = auth0_oauth_token(build_auth0_url(ctx.auth0), code, verifier, ctx.client_id, redirect_uri)

    click.echo("Success!, you sre authorized to use the command line.")

    return data['access_token'], data['refresh_token'], data['id_token']


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


def output_result(ctx, result, fields=None, formatters=None):
    format_tables = ctx.obj.output_format == 'tables'

    if result is None:
        return

    if format_tables:
        result = format_json_data(result, formatters)
        table_data = list(dict_to_csv(result, fields))

        click.echo(PorcelainTable(table_data).table)
    elif hasattr(ctx.obj, 'bad_json_format'):
        print_as_bad_json(result)
    else:
        print_as_json(result)



# coding=utf-8
from .base_db_index import BaseMLIndex
from ..backend_mixin import BackendMixin


class BackendMLIndex(BackendMixin, BaseMLIndex):
    def __init__(self, connection, config, handle_api):
        super(BackendMLIndex, self).__init__(connection, config, handle_api)

    def _create_table_if_needed(self):
        pass

    def set_entries(self, entries):
        if not entries:
            return

        rows = []
        for name, sha, ctime, mtime, mode, uid, gid, size, url in self._decode_entries(entries):
            row = {
                'name': name,
                'sha': sha,
                'ctime': ctime,
                'mtime': mtime,
                'mode': mode,
                'uid': uid,
                'gid': gid,
                'size': size,
                'url': url
            }

            rows.append(row)

        with self._connection.get_cursor() as session:
            url = 'data_volumes/%s/index/stage' % self._volume_id
            msg = {
                'entries': rows,
            }

            self._handle_api(self._config, session.post, url, msg)

    def get_commit_id(self):
        with self._connection.get_cursor() as session:
            url = 'data_volumes/%s/index/stage/commit_id' % self._volume_id
            result = self._handle_api(self._config, session.get, url)

            return result['commit_id']

    def begin_commit(self, tree_id):
        raise NotImplementedError(self.begin_commit)

    def end_commit(self):
        raise NotImplementedError(self.end_commit)



# coding=utf-8
from tests.base import BaseTest
from mali_commands.utilities.tables import format_json_data


class TestTables(BaseTest):

    def testFormatJsonData_formatSpecifiedField(self):
        json_data = [
            {
                'a': 3,
                'b': 5,
            }
        ]

        formatters = {
            'a': lambda x: x * x
        }

        formatted_data = format_json_data(json_data, formatters=formatters)

        row = formatted_data[0]
        self.assertEqual(row['a'], 9)
        self.assertEqual(row['b'], 5)

    def testFormatJsonData_ignoreNonExistentField(self):
        json_data = [
            {
                'a': 3,
                'b': 5,
            }
        ]

        formatters = {
            'c': lambda x: x * x
        }

        formatted_data = format_json_data(json_data, formatters=formatters)
        self.assertEqual(formatted_data, json_data)



# coding=utf-8
import tensorflow as tf

from click.testing import CliRunner

from tests.base import BaseCliTest
from mali import cli
from mali_commands.models import get_tensorflow_weights_hash


class TestModelsWithTensorFlow(BaseCliTest):
    def tearDown(self):
        super(BaseCliTest, self).tearDown()

        # HACK: reset tf.Session monkey patching
        from missinglink_kernel.callback import tensorflow_callback

        # Reset all the monkey patching with tf.Session
        if tensorflow_callback.base_init is not None:
            tf.Session.__init__ = tensorflow_callback.base_init

        tensorflow_callback.base_init = None
        tensorflow_callback.base_run = None

    def testGetWeightsHash(self):
        tf.reset_default_graph()

        runner = CliRunner()
        result = runner.invoke(cli, ['models', 'getWeightsHash',
                                     '--framework', 'tf',
                                     '--model', 'tests/assets/tensorflow_model/model.ckpt'], catch_exceptions=False)

        self.assertEqual(result.exit_code, 0)

    def testGetTensorFlowWeightsHash(self):
        tf.reset_default_graph()

        model_filepath = 'tests/assets/tensorflow_model/model.ckpt'
        weights_hash = get_tensorflow_weights_hash(model_filepath)
        self.assertEqual(weights_hash, 'v1_0eadc7f05dc2ffb1bf0b3429dab4f32f27d86c3e')



# coding=utf-8
import fudge
import httpretty

from mali import cli
from .base import BaseCliTest
from click.testing import CliRunner
from fudge.inspector import arg
from contextlib import closing


class TestData(BaseCliTest):
    user_org = "user_org_" + BaseCliTest.some_random_shit()
    email_domain = "domain_%s@%s" % (BaseCliTest.some_random_shit(), BaseCliTest.some_random_shit())

    @httpretty.activate
    # @fudge.patch('mali_commands.commons.handle_api')
    def test_add_no_volume(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["data", "add", '-f', './'], catch_exceptions=True)
        self.assertNotEquals(result.exit_code, 0)

    @httpretty.activate
    # @fudge.patch('mali_commands.commons.handle_api')
    def test_add_commit_no_message(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["data", "add", 'dassdasda', '-f', './', '--commit'], catch_exceptions=True)
        self.assertNotEquals(result.exit_code, 0)

    @httpretty.activate
    def test_add_files_invalid_volume_id(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["data", "add", 'dassdasda', '-f', './', '--commit'], catch_exceptions=True)
        self.assertNotEquals(result.exit_code, 0)

    @httpretty.activate
    @fudge.patch('mali_commands.data._run_add_batches')
    @fudge.patch('mali_commands.data.with_repo')
    def test_add_run_add_batches_call(self, _run_add_batches, with_repo):
        files_path = './'
        runner = CliRunner()
        volume_id = self.some_random_shit_number_int63()
        fo = fudge.patch('mali_commands.data.with_repo')
        _run_add_batches.expects_call().with_matching_args((files_path,), arg.any()).returns(None).times_called(1)
        with_repo.expects_call().with_matching_args(arg.any(), volume_id).returns(fo).times_called(1)
        result = runner.invoke(cli, ["data", "add", str(volume_id), '-f', files_path], catch_exceptions=False)
        self.assertEquals(result.exit_code, 0)

    @httpretty.activate
    def test_enumerate_paths(self):
        import os
        import tempfile
        file_count = 100

        def make_temp_files(count):
            for i in range(count):
                f = tempfile.NamedTemporaryFile(suffix='.txt', prefix='test_enumerate_paths__', delete=False)
                f.write(bytearray("Hello World!\n{}\n".format(i), 'utf8'))
                f.close()
                yield f.name

        tempfiles = list(make_temp_files(file_count))
        from mali_commands.data import enumerate_paths
        found_files = 0
        for path in enumerate_paths(tempfiles):
            self.assertIn(path, tempfiles)
            found_files += 1
        self.assertEqual(found_files, file_count)

    @httpretty.activate
    def test_enumerate_404_paths(self):
        import os
        import tempfile
        file_count = 100

        def make_temp_files(count):
            for i in range(count):
                f = tempfile.NamedTemporaryFile(suffix='.txt', prefix='test_enumerate_paths__', delete=True)
                # f.write(bytearray("Hello World!\n{}\n".format(i)))
                f.close()
                yield f.name

        tempfiles = list(make_temp_files(file_count))
        from mali_commands.data import enumerate_paths
        found_files = 0
        with self.assertRaises(SystemExit):
            for path in enumerate_paths(tempfiles):
                self.assertIn(path, tempfiles)
                found_files += 1

    @httpretty.activate
    @fudge.patch('mali_commands.data.enumerate_paths')
    # @fudge.patch('mali_commands.data.get_batch_of_files_from_paths')
    @fudge.patch('mali_commands.legit.dulwich.porcelain.add')
    def test_add_run_add_batches_call(self, enumerate_paths, add):
        from mali_commands.data import _run_add_batches

        files_path = ('./',)
        total_files = 1000
        repo = self.get_obj(data_volume_config=self.get_obj(embedded=True))
        enumerate_paths.expects_call().with_matching_args(files_path).returns(list(range(total_files)))

        # get_batch_of_files_from_paths.expects_call().with_matching_args(total_files, 100).returns(list(range(100)))
        closure_dict = {'LastBatch': False}

        def is_valid(x):
            closure_dict['max_value'] = x[-1]
            if len(x) == 10:
                return True
            elif not closure_dict['LastBatch']:
                closure_dict['LastBatch'] = True
                return True
            return False

        add.expects_call().with_matching_args(
            repo,
            arg.passes_test(is_valid),
            # arg.any(),
            True
        ).returns({})
        _run_add_batches(files_path, repo)
        self.assertEqual(total_files - 1, closure_dict['max_value'])



# coding=utf-8
import random
import sys
from .base_connection import BaseConnection


class SqliteSqlHelper(object):
    def __init__(self):
        pass

    @classmethod
    def escape(cls, name):
        return '[%s]' % name

    @classmethod
    def random_function_name(cls):
        return 'ml_random'


class SqliteConnection(BaseConnection):
    def __init__(self, data_volume_config, **kwargs):
        self.__rnd = None
        super(SqliteConnection, self).__init__(data_volume_config, **kwargs)

    def _create_cursor(self):
        return self._native_conn.cursor()

    def _commit(self):
        self._native_conn.commit()

    def _rollback(self):
        self._native_conn.rollback()

    @classmethod
    def execute_multi(cls, cursor, multi_sql):
        for sql in multi_sql.split(';'):
            sql = sql.strip()
            if len(sql) == 0:
                continue

            cursor.execute(sql)

    def _create_connection(self, path=None, read_only=False, **kwargs):
        import sqlite3

        def dict_factory(cursor, row):
            d = {}
            for idx, col in enumerate(cursor.description):
                d[col[0]] = row[idx]

            return d

        def ml_random(seed):
            if self.__rnd is None:
                self.__rnd = random.Random(seed)

            return self.__rnd.random()

        if read_only and sys.version_info >= (3, 4):
            conn = sqlite3.connect('file:{path}?mode=ro'.format(path=path), uri=True)
        else:
            conn = sqlite3.connect(path)

        conn.row_factory = dict_factory
        conn.create_function("ml_random", 1, ml_random)

        return conn

    def create_sql_helper(self):
        return SqliteSqlHelper()



# coding=utf-8
import logging
import six

from flatten_json import unflatten
from .base_metadata_db import BaseMetadataDB
from flatten_json import flatten

logger = logging.getLogger(__name__)


# noinspection SqlIdentifier,SqlResolve
class MySqlMetadataDB(BaseMetadataDB):
    create_table_sql = """
      BEGIN;
      CREATE TABLE IF NOT EXISTS `metadata` (
        `@sha` VARCHAR(128) NOT NULL,
        `@commit_sha` VARCHAR(128) NOT NULL,
        `@version` INT UNSIGNED NOT NULL,
        PRIMARY KEY (`@sha`, `@commit_sha`),
        INDEX `@idx_commit_sha` (`@commit_sha`),
        INDEX `@idx_version` (`@version`)
      );
      CREATE TABLE IF NOT EXISTS `versions` (
        `table` VARCHAR(128) NOT NULL,
        `version` INT UNSIGNED NOT NULL,
        PRIMARY KEY (`table`)
      );
      INSERT INTO `versions` (`table`, `version`) SELECT 'metadata', 0
        WHERE NOT EXISTS(SELECT 1 FROM `versions` WHERE `table` = 'metadata');
      COMMIT;
    """

    def __init__(self, connection):
        self.__prev_table_info = None
        super(MySqlMetadataDB, self).__init__(connection)

    def _create_table(self):
        with self._connection.get_cursor() as c:
            for result in c.execute(self.create_table_sql, multi=True):
                logger.info("'%s': %s", result.statement, result)

    @classmethod
    def __dict_gen_all(cls, cursor):
        field_names = [d[0].lower() for d in cursor.description]
        rows = cursor.fetchall()
        if not rows:
            return

        for row in rows:
            yield dict(zip(field_names, row))

    @classmethod
    def __dict_gen_one(cls, cursor):
        field_names = [d[0].lower() for d in cursor.description]
        row = cursor.fetchone()
        if not row:
            return

        return dict(zip(field_names, row))

    def __get_table_info(self):
        with self._connection.get_cursor() as c:
            c.execute("describe metadata")

            result = {}
            for column_info in self.__dict_gen_all(c):
                # noinspection PyTypeChecker
                result[column_info['field']] = column_info['type']

            return result

    def _add_missing_columns(self, data_object):
        flatten_data = flatten(data_object, separator='.')

        if self.__prev_table_info is None:
            table_info = self.__get_table_info()
            self.__prev_table_info = table_info

        self.__add_missing_columns_from_table_info(flatten_data)

    def __add_missing_columns_from_table_info(self, flatten_data):
        for column_name, column_value in flatten_data.items():
            if column_name not in self.__prev_table_info:
                self.__add_column(column_name, column_value)

    def __add_column(self, name, column_value):
        column_sql_type = self.__value_to_sql_type(column_value)

        with self._connection.get_cursor() as c:
            add_column = "ALTER TABLE metadata ADD COLUMN `{name}` {type}".format(name=name, type=column_sql_type)
            c.execute(add_column)

            add_index = "CREATE INDEX `idx_{name}` ON `metadata` (`{name}`)".format(name=name)
            c.execute(add_index)
            self._connection.commit()

            self.__prev_table_info = self.__get_table_info()

    @classmethod
    def __value_to_sql_type(cls, column_value):
        if isinstance(column_value, six.string_types):
            return 'VARCHAR(128)'

        if isinstance(column_value, six.integer_types):
            return 'INTEGER'

        if isinstance(column_value, (float, )):
            return 'DOUBLE'

        raise Exception('UNKNOWN TYPE %s' % type(column_value))

    def _query_head_data(self, sha_list):
        format_strings = ','.join(['%s'] * len(sha_list))

        get_row_sql = """
            select metadata.*
            from metadata
            inner join (
                SELECT `@sha` as `@__max_sha`, MAX(`@version`) as `_@max_version`
                FROM `metadata`
                WHERE `@sha` in (%s)
                GROUP BY `@sha`) _max
            on `@sha` = `@__max_sha`
            where `@version` = `_@max_version`;
        """ % format_strings

        with self._connection.get_cursor() as c:
            c.execute(get_row_sql, tuple(sha_list))
            return list(self.__dict_gen_all(c))

    def _add_data(self, flatten_data_list):
        fields = self.__prev_table_info.keys()

        values_params = ['%s'] * len(fields)

        values_list = []
        for flatten_data in flatten_data_list:
            values = []

            for field_name in fields:
                values.append(flatten_data.get(field_name))

            values_list.append(values)

        escape_fields = ['`%s`' % field_name for field_name in fields]

        insert_data_sql = "REPLACE INTO `metadata` ({fields}) values ({values_params})".format(
            fields=",".join(escape_fields), values_params=",".join(values_params))

        with self._connection.get_cursor() as c:
            c.executemany(insert_data_sql, values_list)

            self._connection.commit()

    def commit(self, commit_sha, tree_id):
        update_staging = """
          UPDATE `versions` SET `version` = `version` + 1 where `table` = 'metadata';
          UPDATE `metadata` SET `@commit_sha` = '{commit_sha}', `@version` = (SELECT `version` from `versions` where `table` = 'metadata') WHERE `@commit_sha` = 'staging';
        """

        update_staging = update_staging.format(commit_sha=commit_sha)

        with self._connection.get_cursor() as c:
            res = list(c.execute(update_staging, multi=True))

            self._connection.commit()

    def _query(self, sql_vars, select_fields, where):
        query_sql = """
             select {select}
             from `metadata`
             where `@version` IN (
               SELECT MAX(`@version`)
               FROM `metadata`
               GROUP BY `@sha`
             ) AND {where}
             ORDER BY `@sha`
         """.format(where=where, select=','.join(select_fields))

        query_sql = self.fill_in_vars(query_sql, sql_vars)

        for result in self._return_all_result_from_query(query_sql):
            yield result

    def get_data_for_commit(self, sha, commit_sha):
        raise NotImplementedError(self.get_data_for_commit)

    def get_all_data(self, sha):
        raise NotImplementedError(self.get_all_data)



# coding=utf-8
import logging

from google.appengine.ext import ndb

from ..gae_datastore_mixin import GAEDatastoreMixin
from ..dulwich.refs import RefsContainer


SYMREF = b'ref: '


class DataVolumeRefModel(ndb.Model):
    data = ndb.StringProperty(required=True)
    volume_id = ndb.IntegerProperty()  # This is a new prop its needs to be required but for backward support it can't

    @classmethod
    def _get_kind(cls):
        return 'Ref'


class GAEDatastoreRefContainer(GAEDatastoreMixin, RefsContainer):
    def __init__(self, org_id, volume_id):
        super(GAEDatastoreRefContainer, self).__init__(org_id, volume_id)

    def close(self):
        pass

    def delete_all(self):
        self.delete_all_by_class(DataVolumeRefModel)

    @property
    def _entity_kind(self):
        return 'Ref'

    def get_packed_refs(self):
        return {}

    def __add_or_set(self, name, on_ref_entity):
        logging.info('__add_or_set %s', name)

        self._check_refname(name)

        name = name.decode('ascii')

        @ndb.transactional()
        def txn():
            key = self._build_key(name)
            ref_entity = key.get()

            return on_ref_entity(key, ref_entity)

        return txn()

    def set_if_equals(self, name, old_ref, new_ref):
        logging.debug('set_if_equals %s %s %s', name, old_ref, new_ref)

        old_ref = old_ref.decode('ascii')
        new_ref = new_ref.decode('ascii')

        def on_ref_entity(key, ref_entity):
            if ref_entity is None:
                return False

            if ref_entity.data != old_ref:
                return False

            ref_entity.data = new_ref
            ref_entity.put()

            return True

        return self.__add_or_set(name, on_ref_entity)

    def add_if_new(self, name, ref):
        logging.debug('add_if_new %s %s', name, ref)

        ref = ref.decode('ascii')

        def on_ref_entity(key, ref_entity):
            if ref_entity is not None:
                return False

            DataVolumeRefModel(key=key, data=ref, volume_id=self._volume_id).put()

            return True

        return self.__add_or_set(name, on_ref_entity)

    def read_loose_ref(self, name):
        """Read a reference and return its contents.

        If the reference file a symbolic reference, only read the first line of
        the file. Otherwise, only read the first 40 bytes.

        :param name: the refname to read, relative to refpath
        :return: The contents of the ref file, or None if the file does not
            exist.
        """

        ref_key = self._build_key(name.decode('utf8'))
        ref = ref_key.get()

        if ref is None:
            return None

        data = ref.data

        header = data[:len(SYMREF)]
        if header == SYMREF:
            # Read only the first line
            return data.splitlines()[0]

        # Read only the first 40 bytes
        return data[:40].encode('ascii')

    def set_symbolic_ref(self, name, other):
        raise NotImplementedError(self.set_symbolic_ref)

    def allkeys(self):
        raise NotImplementedError(self.allkeys)

    def remove_if_equals(self, name, old_ref):
        raise NotImplementedError(self.remove_if_equals)



# coding=utf-8
import logging
import multiprocessing
import uuid
from threading import Semaphore
from multiprocessing import Pool
import os
import signal


def worker_init(parent_id):
    def sig_int(signal_num, frame):
        os.kill(parent_id, signal.SIGSTOP)

    signal.signal(signal.SIGINT, sig_int)


class MethodWrapper(object):
    def __init__(self, callable_func):
        self.__callable = callable_func

    def __call__(self, *args, **kwargs):
        try:
            self.__callable(*args, **kwargs)
        except KeyboardInterrupt as e:
            pass


class MultiProcessControlShim(object):
    @classmethod
    def execute(cls, method, args, callback=None):
        method(*args)
        if callback:
            callback(None)

    def terminate(self):
        pass

    def close(self):
        pass


class MultiProcessControl(object):
    def __init__(self, processes=-1):
        processes = multiprocessing.cpu_count() * 2 if processes == -1 else processes
        self.__upload_pool = Pool(processes, worker_init, initargs=(os.getpid(), ))
        self.__max_waiting_semaphore = Semaphore(processes * 4)
        self.__jobs = {}

    def close(self):
        if self.__upload_pool is None:
            return

        logging.debug('%s closing & joining pool', self.__class__)
        self.__upload_pool.close()
        self.__upload_pool.join()
        self.__check_pending_jobs()
        logging.debug('%s pool completed', self.__class__)

    def create_callback(self, token, callback):
        def do_callback(param):
            self.__max_waiting_semaphore.release()
            if callback is not None:
                callback(param)

        return do_callback

    def __check_pending_jobs(self):
        for token, async_result in self.__jobs.items():
            if async_result is None:
                continue

            if not async_result.ready():
                continue

            async_result.get()
            self.__jobs[token] = None

        self.__jobs = {token: async_result for token, async_result in self.__jobs.items() if async_result is not None}

    def execute(self, method, args, callback=None):
        self.__max_waiting_semaphore.acquire()

        self.__check_pending_jobs()

        token = uuid.uuid4()
        job_async_result = self.__upload_pool.apply_async(
            MethodWrapper(method),
            args=args,
            callback=self.create_callback(token, callback))

        self.__jobs[token] = job_async_result

    def terminate(self):
        self.__upload_pool.terminate()



# coding=utf-8
from google.appengine.ext import ndb


class GAEDatastoreMixin(object):
    def __init__(self, org_id, volume_id):
        self.__volume_id = volume_id
        self.__org_id = org_id

    def delete_all_by_class(self, klass):
        q = klass.query(
            klass.volume_id == self._volume_id, namespace=self._org_id)

        while True:
            refs = list(q.fetch(100, keys_only=True))
            if len(refs) == 0:
                return

            ndb.delete_multi(refs)

    @property
    def _volume_id(self):
        return self.__volume_id

    @property
    def _org_id(self):
        return self.__org_id

    @classmethod
    def build_key_from_volume_id(cls, org_id, volume_id, kind, name):
        parent_key = ndb.Key('DataVolume', volume_id, namespace=org_id)

        name = name or 1

        return ndb.Key(kind, name, parent=parent_key, namespace=org_id)

    def _build_key(self, name):
        return self.build_key_from_volume_id(self.__org_id, self.__volume_id, self._entity_kind, name)

    @property
    def _entity_kind(self):
        raise NotImplementedError(self._entity_kind)



# coding=utf-8
import os
from setuptools import setup, find_packages

build = os.environ.get('PIP_BUILD', 100)
keywords = os.environ.get('PIP_KEYWORDS', 'test')


version = '0.{}'.format(build)

with open('requirements.txt') as f:
    install_requires = [r for r in f.readlines() if len(r.strip()) > 0]

setup(
    name='mali',
    version=version,
    description='Command line tool for missinglink.ai platform',
    author='missinglink.ai',
    author_email='support+mali@missinglink.ai',
    url='https://missinglink.ai',
    license='Apache',
    py_modules=['mali'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 2.7'
    ],
    packages=find_packages(),
    install_requires=install_requires,
    entry_points='''
        [console_scripts]
        mali=mali:main
    ''',
)
